<?php

namespace Drupal\ino_pt_list\Plugin\Field\FieldFormatter;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\taxonomy\TermInterface;
use Drupal\views\Element\View;
use Drupal\views\ViewExecutable;
use Drupal\views\Views;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class ListFormatter.
 *
 * @FieldFormatter(
 *   id = "ino_pt_list_formatter",
 *   label = @Translation("Cappuccino content collection node formatter"),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 *
 * @package Drupal\ino_pt_list\Plugin\Field\FieldFormatter
 */
class ListFormatter extends FormatterBase {

  // Default values for the node entity type.
  // They can vary in the constructor for other entity types.
  protected $entity_type = 'node';
  protected $view_id = 'ino_pt_list_nodes';
  protected $items_field = 'field_ph_list_items';
  protected $content_type_field = 'field_ph_list_content_type';

  const LAYOUT_FIELD = 'field_ph_list_layout';
  const PROMOTED_FIELD = 'field_ph_list_promoted';
  const CURRENT_TERM_FIELD = 'field_ph_list_current_term';
  const TERM_FILTERS_FIELD = 'field_ph_list_term_filters';
  const COUNT_FIELD = 'field_ph_list_count';
  const PAGER_FIELD = 'field_ph_list_pager';
  const SELECTION_TYPE_FIELD = 'field_ph_list_type';

  const INO_PT_LIST_MAX_ITEMS = 10000;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The current route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $currentRouteMatch;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, EntityTypeManagerInterface $entityTypeManager, LanguageManagerInterface $languageManager, EntityFieldManagerInterface $entityFieldManager, RouteMatchInterface $currentRouteMatch) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->entityTypeManager = $entityTypeManager;
    $this->languageManager = $languageManager;
    $this->entityFieldManager = $entityFieldManager;
    $this->currentRouteMatch = $currentRouteMatch;

    if ($this->getSetting('entity_type') === 'term') {
      $this->entity_type = 'taxonomy_term';
      $this->view_id = 'ino_pt_list_terms';
      $this->items_field = 'field_ph_list_taxonomy_items';
      $this->content_type_field = 'field_ph_list_taxonomy_type';
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('entity_type.manager'),
      $container->get('language_manager'),
      $container->get('entity_field.manager'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    return [$this->t('Entity type: @entity_type', ['@entity_type' => $this->getSetting('entity_type')])];
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'entity_type' => 'node',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form['entity_type'] = [
      '#title' => $this->t('Entity Type'),
      '#type' => 'select',
      '#options' => [
        'node' => $this->t('Node'),
        'term' => $this->t('Taxonomy Term'),
      ],
      '#default_value' => $this->getSetting('entity_type') ?? 'node',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function view(FieldItemListInterface $items, $langcode = NULL) {
    // Default the language to the current content language.
    if (empty($langcode)) {
      $langcode = $this->languageManager->getCurrentLanguage(LanguageInterface::TYPE_CONTENT)->getId();
    }
    return $this->viewElements($items, $langcode);
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $content_collection_paragraph = $items->getParent()->getValue();
    if (!($content_collection_paragraph instanceof ParagraphInterface && $content_collection_paragraph->getType() === 'ino_pt_list')) {
      return [];
    }
    return $this->prepareItems($content_collection_paragraph);
  }

  /**
   * Handles the content collection logic.
   *
   * @param \Drupal\paragraphs\ParagraphInterface $content_collection_paragraph
   *   The content collection paragraph.
   *
   * @return array|array[]
   *   The items to be rendered.
   */
  protected function prepareItems(ParagraphInterface $content_collection_paragraph) {
    // Layouts are views displays.
    $display_id = $content_collection_paragraph->get(static::LAYOUT_FIELD)->value;
    // See views_embed_view() and Drupal\views\Plugin\Block\ViewsBlock::build().
    $view = Views::getView($this->view_id);
    if (!$view || !$view->access($display_id)) {
      return [];
    }
    $view->setDisplay($display_id);

    $selection_type = $content_collection_paragraph->get(static::SELECTION_TYPE_FIELD)->value;
    if (($selection_type === 'manual' && !$this->processManualSelection($content_collection_paragraph, $view)) ||
      ($selection_type === 'automatic' && !$this->processAutomaticSelection($content_collection_paragraph, $view))) {
      // The process method returned FALSE => No results.
      return [];
    }

    // Get selected/filtered content.
    $options = [
      'offset' => 0,
      'has_pagination' => (bool) $content_collection_paragraph->get(static::PAGER_FIELD)->value,
      'items_per_page' => static::INO_PT_LIST_MAX_ITEMS,
    ];
    $item_count = $content_collection_paragraph->get(static::COUNT_FIELD)->value;
    if ($item_count && ($options['has_pagination'] || $selection_type === 'automatic')) {
      // Manual selection has no item count without a pager.
      $options['items_per_page'] = $item_count;
    }
    // Items per page / Pager.
    static $pager_count;
    if (!isset($pager_count)) {
      $pager_count = 0;
    }
    $pager = $view->display_handler->getOption('pager');
    $pager['options']['items_per_page'] = $options['items_per_page'];
    // Avoid pager conflicts when having multiple content collections on a page.
    $pager['options']['id'] = $pager_count++;
    $pager['options']['offset'] = $options['offset'];
    // Add pager or just limit the number of items.
    $pager['type'] = $options['has_pagination'] ? 'full' : 'some';
    $view->display_handler->setOption('pager', $pager);

    // Build the render array.
    $args = [];
    // Exclude the current node/term (with a views argument).
    if ($reference_entity = $this->getParagraphParentEntity($content_collection_paragraph)) {
      // TODO: check entity type.
      $args[] = $reference_entity->id();
    }
    $output = $view->buildRenderable($display_id, $args, FALSE);

    // Add contextual links.
    $output['#view_id'] = $view->storage->id();
    $output['#view_display_show_admin_links'] = $view->getShowAdminLinks();
    $output['#view_display_plugin_id'] = $view->display_handler->getPluginId();
    views_add_contextual_links($output, 'view', $display_id);
    $output = View::preRenderViewElement($output);
    if (empty($output['view_build'])) {
      $output = ['#cache' => $output['#cache']];
    }
    return $output;
  }

  /**
   * Handles the manual selection logic.
   *
   * @param \Drupal\paragraphs\ParagraphInterface $content_collection_paragraph
   *   The content collection paragraph.
   * @param \Drupal\views\ViewExecutable $view
   *   The corresponding view.
   *
   * @return bool
   *   FALSE if this should return no results. TRUE otherwise.
   */
  protected function processManualSelection(ParagraphInterface $content_collection_paragraph, ViewExecutable $view) {
    // Manual selection - a given list of ids.
    $ids = [];
    $query_base_table = $view->storage->get('base_table');
    $query = $view->getQuery();
    if ($content_collection_paragraph->get($this->items_field)->isEmpty()) {
      // No items added to the manual selection list.
      return FALSE;
    }
    foreach ($content_collection_paragraph->get($this->items_field)->getValue() as $list_item) {
      $ids[$list_item['target_id']] = $list_item['target_id'];
    }
    try {
      $entity_type = $this->entityTypeManager->getDefinition($this->entity_type);
    }
    catch (PluginNotFoundException $ignored) {
      // Shouldn't happen, as node/term storage should always exist.
      return FALSE;
    }
    $id_column = $entity_type->getKeys()['id'];
    // Duplicates are not shown.
    $query->addWhere(0, "$query_base_table.$id_column", $ids, 'IN');
    // Manual sort order.
    // ORDER BY FIELD(ID,1,5,4,3).
    $id_order = implode(',', $ids);
    $formula = "FIELD($query_base_table.$id_column,$id_order)";
    $query->addOrderBy(NULL, $formula, 'ASC', 'ino_pt_list_sort');
    return TRUE;
  }

  /**
   * Handles the automatic selection logic.
   *
   * @param \Drupal\paragraphs\ParagraphInterface $content_collection_paragraph
   *   The content collection paragraph.
   * @param \Drupal\views\ViewExecutable $view
   *   The corresponding view.
   *
   * @return bool
   *   FALSE if this should return no results. TRUE otherwise.
   */
  protected function processAutomaticSelection(ParagraphInterface $content_collection_paragraph, ViewExecutable $view) {
    // Automatic selection - set views filters.
    $bundles = [];
    $query_base_table = $view->storage->get('base_table');
    $type_items = $content_collection_paragraph->get($this->content_type_field)->getValue();
    $display_id = $content_collection_paragraph->get(static::LAYOUT_FIELD)->value;
    foreach ($type_items as $item) {
      $bundles[$item['target_id']] = $item['target_id'];
    }
    if (empty($bundles)) {
      // Bundle filter is required, return no results.
      return FALSE;
    }
    $filter_options = [
      'query_base_table' => $query_base_table,
      'entity_type_id' => $this->entity_type,
      'bundles' => $bundles,
      'display_id' => $display_id,
      'filter_promoted' => FALSE,
    ];
    // 'Promoted to front page' filter.
    $promoted = $content_collection_paragraph->get(static::PROMOTED_FIELD)->getValue();
    $filter_options['filter_promoted'] = isset($promoted[0]['value']) && $promoted[0]['value'];
    if ($content_collection_paragraph->get(static::CURRENT_TERM_FIELD)->value == 1) {
      $reference_entity = $this->getParagraphParentEntity($content_collection_paragraph);
      if ($reference_entity !== NULL && $reference_entity instanceof TermInterface) {
        $term_filters_value[0]['settings'] = [
          $reference_entity->bundle() => [
            'use_global' => 0,
            'term_reference' => [['target_id' => $reference_entity->id()]],
          ],
        ];
      }
    }
    else {
      $term_filters_value = $content_collection_paragraph->get(static::TERM_FILTERS_FIELD)->getValue();
    }
    $filter_options['term_filters'] = [];
    if (isset($reference_entity, $term_filters_value[0])) {
      $filter_options['term_filters'] = $this->getTagFilters($term_filters_value[0]['settings'], $reference_entity);
    }
    elseif (isset($term_filters_value[0])) {
      $filter_options['term_filters'] = $this->getTagFilters($term_filters_value[0]['settings']);
    }
    $bundles = $this->addViewFilters($view, $filter_options);
    if (empty($bundles)) {
      // The view should return no results.
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Gets a paragraph's parent or current page entity.
   *
   * @param \Drupal\paragraphs\ParagraphInterface $paragraph
   *   The paragraph object.
   *
   * @return \Drupal\Core\Entity\ContentEntityBase|null
   *   The reference entity.
   */
  protected function getParagraphParentEntity(ParagraphInterface $paragraph) {
    $reference_entity = NULL;
    $parent_entity = $paragraph->getParentEntity();
    if (!empty($parent_entity)) {
      switch ($parent_entity->getEntityTypeId()) {
        case 'node':
          // Paragraph is on a node.
          $reference_entity = $parent_entity;
          break;

        case 'config_pages':
          // TODO: Looking for a node or a term.
          // Global content, look for the current node/entity.
          $reference_entity = $this->currentRouteMatch->getParameter('node');
          break;

        case 'taxonomy_term':
          $reference_entity = $parent_entity;
          break;

        case 'paragraph':
          $reference_entity = $this->getParagraphParentEntity($parent_entity);
          break;
      }
    }

    return $reference_entity;
  }

  /**
   * Loads the list of tags used to filter content.
   *
   * @param array $term_filter_settings
   *   The settings value of the term_reference_filter field, that needs to be
   *   parsed and terms loaded according to the specifications.
   * @param \Drupal\Core\Entity\FieldableEntityInterface | NULL $entity
   *   The parent entity.
   *
   * @return array
   *   Array of taxonomy terms.
   */
  protected function getTagFilters(array $term_filter_settings, FieldableEntityInterface $entity = NULL) {
    $terms = [];
    foreach ($term_filter_settings as $vid => $term_filter) {
      $tids = [];
      if ($term_filter['use_global'] === 0) {
        if (!empty($term_filter['term_reference'])) {
          foreach ($term_filter['term_reference'] as $tid) {
            $tids[] = $tid['target_id'];
          }
        }
      }
      elseif (!empty($entity)) {
        // Load terms from the related entity's fields
        // instead of the paragraph's.
        $field_name_default = 'field_n_' . $vid;
        $field_name_custom = 'field_' . INO_CONTENT_EDITING_FIELD_NAME_PREFIX . 'n_' . $vid;
        $field_name = $entity->hasField($field_name_default)
          ? $field_name_default
          : ($entity->hasField($field_name_custom) ? $field_name_custom : '');
        if ($field_name) {
          $referenced_terms = $entity->get($field_name)->referencedEntities();
          foreach ($referenced_terms as $referenced_term) {
            $tids[] = $referenced_term->id();
          }
        }
      }
      if (!empty($tids)) {
        $terms[$vid] = $tids;
      }
    }
    return $terms;
  }

  /**
   * Adds view filters for automatic selection.
   *
   * @param \Drupal\views\ViewExecutable $view
   *   The view.
   * @param array $options
   *   Filter options/values.
   *
   * @return array
   *   The bundles tht we are filtering for.
   */
  protected function addViewFilters(ViewExecutable $view, array $options) : array {
    $query_base_table = $options['query_base_table'];
    $entity_type_id = $options['entity_type_id'];
    $bundles = $options['bundles'];
    $display_id = $options['display_id'];
    $filter_promoted = ('node' == $entity_type_id) ? $options['filter_promoted'] : FALSE;
    $term_filters = $options['term_filters'];
    // Add filters into the default filter group.
    $filter_group_id = 1;
    // Filter by bundle (content type / vocabulary).
    $field = 'node' == $entity_type_id ? 'type' : 'vid';
    $options = [
      'relationship' => 'none',
      'group_type' => 'group',
      'group' => $filter_group_id,
      'admin_label' => '',
      'operator' => 'in',
      'value' => empty($term_filters) ? $bundles : [],
      'entity_type' => $entity_type_id,
      'entity_field' => $field,
      'plugin_id' => 'bundle',
    ];
    $view->addHandler($display_id, 'filter', $query_base_table, $field, $options);
    // 'Promoted' filter.
    if ('node' == $entity_type_id && $filter_promoted) {
      $options = [
        'relationship' => 'none',
        'group_type' => 'group',
        'group' => $filter_group_id,
        'admin_label' => '',
        'operator' => '=',
        'value' => '1',
        'entity_type' => 'node',
        'entity_field' => 'promote',
        'plugin_id' => 'boolean',
      ];
      $view->addHandler($display_id, 'filter', $query_base_table, 'promote', $options);
    }
    // Taxonomy term filters.
    if (!empty($term_filters)) {
      // Add term filters for each bundle.
      if (count($bundles) > 1) {
        // Terms per bundle in 'OR' filter groups.
        // We already have published, language, bundle and promoted filters.
        // Views filter groups cannot be nested,
        // repeat default filters per group.
        $filter_groups = $view->display_handler->getOption('filter_groups');
        $filter_groups['operator'] = 'OR';
        $view->display_handler->overrideOption('filter_groups', $filter_groups);
      }
      $filters_default = $view->display_handler->getOption('filters');
      $term_ref_per_bundle = $this->getTermReferenceFieldsForBundle($bundles, $entity_type_id);
      // Views filter groups cannot be nested.
      // Add separate filter groups for each bundle.
      foreach (array_values($bundles) as $bundle) {
        // Get vids not referenced by the bundle.
        $missing_vids = array_diff_key($term_filters, $term_ref_per_bundle[$bundle]);
        if (!empty($missing_vids)) {
          // Exclude current bundle.
          unset($bundles[$bundle]);
          continue;
        }
        foreach ($term_filters as $vid => $tids) {
          foreach ($term_ref_per_bundle[$bundle][$vid] as $field_info) {
            $options = [
              'relationship' => 'none',
              'group_type' => 'group',
              'group' => $filter_group_id,
              'admin_label' => '',
              'operator' => 'or',
              'value' => $tids,
              'reduce_duplicates' => TRUE,
              'type' => 'textfield',
              'limit' => TRUE,
              'vid' => $vid,
              'hierarchy' => FALSE,
              'plugin_id' => 'taxonomy_index_tid',
            ];
            $view->addHandler($display_id, 'filter', $field_info['table'], $field_info['field'], $options);
          }
        }
        // Other filters for this filter group.
        if (1 == $filter_group_id) {
          // Use the first filter group for the first bundle.
          $filters = $view->display_handler->getOption('filters');
          $filters[$field]['value'] = [$bundle => $bundle];
          $view->display_handler->overrideOption('filters', $filters);
        }
        else {
          // Add the default filters.
          // If the group does not yet exist it will be created as an AND group.
          foreach ($filters_default as $filter) {
            if ('type' == $filter['id']) {
              $filter['value'] = [$bundle => $bundle];
            }
            unset($filter['id']);
            $filter['group'] = $filter_group_id;
            $view->addHandler($display_id, 'filter', $filter['table'], $filter['field'], $filter);
          }
        }
        $filter_group_id++;
      }
    }

    return $bundles;
  }

  /**
   * Gets term reference field details in the given bundles.
   *
   * This function presumes that field names are the same.
   *
   * @param array $bundles
   *   The bundle list.
   * @param string $entity_type_id
   *   The entity type ID. Only entity types that implement
   *   \Drupal\Core\Entity\FieldableEntityInterface are supported.
   *
   * @return array
   *   The array of term fields, keyed by vocabulary.
   */
  protected function getTermReferenceFieldsForBundle(array $bundles, $entity_type_id = 'node') : array {
    $vocab_field_info = [];
    // Get an instance of Drupal\Core\Entity\Sql\DefaultTableMapping class.
    $table_mapping = $this->entityTypeManager->getStorage($entity_type_id)->getTableMapping();
    /** @var \Drupal\Core\Entity\Sql\DefaultTableMapping $table_mapping */
    foreach ($bundles as $bundle) {
      if ('_none' == $bundle) {
        continue;
      }
      $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
      foreach ($fields as $field_name => $field_item) {
        /** @var \Drupal\field\Entity\FieldConfig $field_item */
        if ('entity_reference' == $field_item->getType() && !empty($field_item->getTargetBundle())) {
          $settings = $field_item->getSettings();
          if (isset($settings['target_type']) && 'taxonomy_term' == $settings['target_type']) {
            $storage_definition = $field_item->getFieldStorageDefinition();
            $data_table = $table_mapping->getDedicatedDataTableName($storage_definition);
            // E.g. 'target_id'.
            $value_property = key($storage_definition->getSchema()['columns']);
            $field_column_name = $table_mapping->getFieldColumnName($storage_definition, $value_property);
            foreach ($settings['handler_settings']['target_bundles'] as $vid) {
              $vocab_field_info[$bundle][$vid][$field_name] = [
                // E.g. 'node__field_n_tags'.
                'table' => $data_table,
                // E.g. 'field_n_tags_target_id'.
                'field' => $field_column_name,
              ];
            }
          }
        }
      }
    }

    return $vocab_field_info;
  }

}
